from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType, TooltipMode
from sw_mc_lib.XMLParser import XMLParserElement


class TooltipBoolean(Component):
    """
    Displays an on/off signal on this microcontroller's tooltip when it is looked at by the player
    on a spawned vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        label_property: str = "value",
        on_label_property: str = "on",
        off_label_property: str = "off",
        mode_property: TooltipMode = TooltipMode.Always,
        display_number_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.TooltipBoolean, component_id, position, 0.75)
        self.display_number_input: Optional[Input] = display_number_input
        self.label_property: str = label_property
        self.mode_property: TooltipMode = mode_property
        self.on_label_property: str = on_label_property
        self.off_label_property: str = off_label_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> TooltipBoolean:
        assert element.tag == "c", f"invalid TooltipBoolean {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.TooltipBoolean.value
        ), f"Not an TooltipBoolean {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = TooltipBoolean._basic_in_parsing(obj)
        label_property: str = obj.attributes.get("l", "")
        on_label_property: str = obj.attributes.get("on", "")
        off_label_property: str = obj.attributes.get("off", "")
        mode_property: TooltipMode = TooltipMode(int(obj.attributes.get("m", "0")))
        return TooltipBoolean(
            component_id,
            position,
            label_property,
            on_label_property,
            off_label_property,
            mode_property,
            inputs.get("1"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {
            "l": self.label_property,
            "on": self.on_label_property,
            "off": self.off_label_property,
        }
        if self.mode_property != TooltipMode.Always:
            attributes["m"] = str(self.mode_property.value)
        return attributes, self._pos_in_to_xml(self.display_number_input)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

from .SubTypes.ValueComponent import ValueComponent


class PropertyNumber(ValueComponent):
    """
    Adds a custom number input that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        name: str = "number",
        value_property: Optional[NumberProperty] = None,
    ):
        super().__init__(
            ComponentType.PropertyNumber, component_id, position, 0.5, value_property
        )
        self.name: str = name

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertyNumber:
        assert element.tag == "c", f"invalid PropertyNumber {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertyNumber.value
        ), f"Not an PropertyNumber {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, properties = PropertyNumber._basic_in_parsing(obj)
        name: str = obj.attributes.get("n", "number")
        value_property = PropertyNumber._basic_value_parsing(properties)
        return PropertyNumber(component_id, position, name, value_property)

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml()
        children.extend(self._value_to_xml())
        return {}, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class XOR(Component):
    """
    Outputs the logical XOR of its two input signals.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        a_input: Optional[Input] = None,
        b_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.XOR, component_id, position, 0.75)
        self.a_input: Optional[Input] = a_input
        self.b_input: Optional[Input] = b_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> XOR:
        assert element.tag == "c", f"invalid XOR {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.XOR.value
        ), f"Not an XOR {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = XOR._basic_in_parsing(obj)
        return XOR(component_id, position, inputs.get("1"), inputs.get("2"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(self.a_input, self.b_input)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class Blinker(Component):
    """
    Outputs a value that blinks between on and off at a set rate.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        control_signal_input: Optional[Input] = None,
        blink_on_duration_property: float = 1.0,
        blink_off_duration_property: float = 1.0,
    ):
        super().__init__(ComponentType.Blinker, component_id, position, 1.0)
        self.control_signal_input: Optional[Input] = control_signal_input
        self.blink_on_duration_property: float = blink_on_duration_property
        self.blink_off_duration_property: float = blink_off_duration_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> Blinker:
        assert element.tag == "c", f"invalid Blinker {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.Blinker.value
        ), f"Not an Blinker {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Blinker._basic_in_parsing(obj)
        blink_on_duration_property: float = float(obj.attributes.get("on", "1"))
        blink_off_duration_property: float = float(obj.attributes.get("off", "1"))
        return Blinker(
            component_id,
            position,
            inputs.get("1"),
            blink_on_duration_property,
            blink_off_duration_property,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {
            "on": str(self.blink_on_duration_property),
            "off": str(self.blink_off_duration_property),
        }
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.control_signal_input
        )
        return attributes, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class CompositeWriteBoolean(Component):
    """
    Writes up to 32 on/off signals to a composite link in a single logic tick.
    Only connected channels will be modified. The number of inputs and channel to begin writing at can be configured.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        start_channel_property: int = 1,
        channel_count_property: int = 1,
        composite_signal_input: Optional[Input] = None,
        start_channel_input: Optional[Input] = None,
        channel_inputs: Optional[dict[int, Input]] = None,
    ):
        height: float = (
            0.5
            + channel_count_property * 0.25
            + (0.25 if start_channel_property == 0 else 0)
        )
        super().__init__(
            ComponentType.CompositeWriteBoolean, component_id, position, height
        )
        self.start_channel_property: int = start_channel_property
        self.channel_count_property: int = channel_count_property
        self.composite_signal_input: Optional[Input] = composite_signal_input
        self.start_channel_input: Optional[Input] = start_channel_input
        self.channel_inputs: dict[int, Input] = channel_inputs or {}

    @staticmethod
    def from_xml(element: XMLParserElement) -> CompositeWriteBoolean:
        assert element.tag == "c", f"invalid CompositeWriteBoolean {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.CompositeWriteBoolean.value
        ), f"Not an CompositeWriteBoolean {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = CompositeWriteBoolean._basic_in_parsing(obj)
        start_channel_property: int = int(obj.attributes.get("offset", "0")) + 1
        channel_count_property: int = int(obj.attributes.get("count", "0"))
        channel_inputs: dict[int, Input] = {
            i: current_input for i in range(32) if (current_input := inputs.get(str(i)))
        }
        return CompositeWriteBoolean(
            component_id,
            position,
            start_channel_property,
            channel_count_property,
            inputs.get("c"),
            inputs.get("off"),
            channel_inputs,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {"count": str(self.channel_count_property)}
        if self.start_channel_property != 1:
            attributes["offset"] = str(self.start_channel_property - 1)
        inputs: list[Optional[Input]] = []
        for i, current_input in self.channel_inputs.items():
            while len(inputs) < i:
                inputs.append(None)
            inputs[i - 1] = current_input
        children: list[XMLParserElement] = self._pos_in_to_xml(
            *inputs,
            named_inputs={
                "c": self.composite_signal_input,
                "off": self.start_channel_input,
            },
        )
        return {}, children



from __future__ import annotations

from abc import ABC, abstractmethod
from itertools import chain
from typing import Optional, Type

from .Input import Input
from .NumberProperty import NumberProperty
from .Position import Position
from .Types import ComponentType
from .XMLElement import XMLElement
from .XMLParser import XMLParserElement

INNER_TO_XML_RESULT = tuple[dict[str, str], list[XMLParserElement]]


class Component(XMLElement, ABC):
    """
    A generic Component
    """

    width: float = 1

    def __init__(
        self,
        component_type: ComponentType,
        component_id: int,
        position: Position,
        height: float,
    ):
        self.type: ComponentType = component_type
        self.component_id: int = component_id
        self.position: Position = position
        self.height: float = height

    @staticmethod
    def _basic_in_parsing(
        element: XMLParserElement,
    ) -> tuple[int, Position, dict[str, Input], dict[str, NumberProperty]]:
        component_id: int = int(element.attributes.get("id", "0"))
        position: Position = Position()
        inputs: dict[str, Input] = {}
        properties: dict[str, NumberProperty] = {}
        for child in element.children:
            if child.tag == "pos":
                position = Position.from_xml(child)
            elif child.tag.startswith("in"):
                input: Input = Input.from_xml(child)
                if input.component_id > 0:
                    inputs[input.index] = input
            else:
                properties[child.tag] = NumberProperty.from_xml(child)
        return component_id, position, inputs, properties

    @abstractmethod
    def _inner_to_xml(self) -> tuple[dict[str, str], list[XMLParserElement]]:
        ...

    def to_xml(self) -> XMLParserElement:
        object_element = XMLParserElement("object", {"id": str(self.component_id)}, [])
        inner_attributes, inner_children = self._inner_to_xml()
        object_element.attributes.update(inner_attributes)
        object_element.children = inner_children
        return XMLParserElement("c", {"type": str(self.type.value)}, [object_element])

    def to_state_xml(self, index: int) -> XMLParserElement:
        """
        Convert this Component to a XML state entity

        :param index: The zero based index of the entity
        :return: XML state representation
        """
        c_element: XMLParserElement = XMLParserElement(
            f"c{index}", {"id": str(self.component_id)}
        )
        inner_attributes, inner_children = self._inner_to_xml()
        c_element.attributes.update(inner_attributes)
        c_element.children = inner_children
        return c_element

    def _pos_in_to_xml(
        self,
        *inputs: Optional[Input],
        named_inputs: Optional[dict[str, Optional[Input]]] = None,
        properties: Optional[dict[str, NumberProperty]] = None,
    ) -> list[XMLParserElement]:
        """
        Turn the position and inputs into an element. Will strip all None inputs. Will also rename all numbered inputs
        to their according supplied index, and all named inputs to the supplied name.

        :param inputs: Array of inputs, the order is important, as it defines overrides
        :param named_inputs: Inputs by name, used for overrides
        :param properties: Properties by name
        :return: XML elements for all relevant children
        """
        named_inputs = named_inputs or {}
        properties = properties or {}
        children: list[XMLParserElement] = [self.position.to_xml()]
        for i, input in chain(enumerate(inputs), named_inputs.items()):
            if input is not None:
                if isinstance(i, int):
                    input.index = str(i + 1)
                else:
                    assert isinstance(i, str)
                    input.index = i
                children.append(input.to_xml())
        for name, property in properties.items():
            property.name = name
            children.append(property.to_xml())
        return children

    @staticmethod
    def from_xml(element: XMLParserElement) -> Component:
        # pylint: disable=too-many-statements
        component_type: ComponentType = ComponentType(
            int(element.attributes.get("type", "0"))
        )
        element_class: Optional[Type[Component]] = None
        match component_type:
            case ComponentType.Abs:
                element_class = Abs
            case ComponentType.Add:
                element_class = Add
            case ComponentType.AND:
                element_class = AND
            case ComponentType.ArithmeticFunction1In:
                element_class = ArithmeticFunction1In
            case ComponentType.ArithmeticFunction3In:
                element_class = ArithmeticFunction3In
            case ComponentType.ArithmeticFunction8In:
                element_class = ArithmeticFunction8In
            case ComponentType.AudioSwitchbox:
                element_class = AudioSwitchbox
            case ComponentType.Blinker:
                element_class = Blinker
            case ComponentType.BooleanFunction4In:
                element_class = BooleanFunction4In
            case ComponentType.BooleanFunction8In:
                element_class = BooleanFunction8In
            case ComponentType.Capacitor:
                element_class = Capacitor
            case ComponentType.Clamp:
                element_class = Clamp
            case ComponentType.CompositeBinaryToNumber:
                element_class = CompositeBinaryToNumber
            case ComponentType.CompositeReadBoolean:
                element_class = CompositeReadBoolean
            case ComponentType.CompositeReadNumber:
                element_class = CompositeReadNumber
            case ComponentType.CompositeSwitchbox:
                element_class = CompositeSwitchbox
            case ComponentType.CompositeWriteBoolean:
                element_class = CompositeWriteBoolean
            case ComponentType.CompositeWriteNumber:
                element_class = CompositeWriteNumber
            case ComponentType.ConstantNumber:
                element_class = ConstantNumber
            case ComponentType.ConstantOn:
                element_class = ConstantOn
            case ComponentType.Delta:
                element_class = Delta
            case ComponentType.Divide:
                element_class = Divide
            case ComponentType.Equal:
                element_class = Equal
            case ComponentType.GreaterThan:
                element_class = GreaterThan
            case ComponentType.JKFlipFlop:
                element_class = JKFlipFlop
            case ComponentType.LessThan:
                element_class = LessThan
            case ComponentType.LuaScript:
                element_class = LuaScript
            case ComponentType.MemoryRegister:
                element_class = MemoryRegister
            case ComponentType.Modulo:
                element_class = Modulo
            case ComponentType.Multiply:
                element_class = Multiply
            case ComponentType.NAND:
                element_class = NAND
            case ComponentType.NOR:
                element_class = NOR
            case ComponentType.NOT:
                element_class = NOT
            case ComponentType.NumberToCompositeBinary:
                element_class = NumberToCompositeBinary
            case ComponentType.NumericalJunction:
                element_class = NumericalJunction
            case ComponentType.NumericalSwitchbox:
                element_class = NumericalSwitchbox
            case ComponentType.OR:
                element_class = OR
            case ComponentType.PIDController:
                element_class = PIDController
            case ComponentType.PIDControllerAdvanced:
                element_class = PIDControllerAdvanced
            case ComponentType.PropertyDropdown:
                element_class = PropertyDropdown
            case ComponentType.PropertyNumber:
                element_class = PropertyNumber
            case ComponentType.PropertySlider:
                element_class = PropertySlider
            case ComponentType.PropertyText:
                element_class = PropertyText
            case ComponentType.PropertyToggle:
                element_class = PropertyToggle
            case ComponentType.Pulse:
                element_class = Pulse
            case ComponentType.PushToToggle:
                element_class = PushToToggle
            case ComponentType.SRLatch:
                element_class = SRLatch
            case ComponentType.Subtract:
                element_class = Subtract
            case ComponentType.Threshold:
                element_class = Threshold
            case ComponentType.TimerRTF:
                element_class = TimerRTF
            case ComponentType.TimerRTO:
                element_class = TimerRTO
            case ComponentType.TimerTOF:
                element_class = TimerTOF
            case ComponentType.TimerTON:
                element_class = TimerTON
            case ComponentType.TooltipBoolean:
                element_class = TooltipBoolean
            case ComponentType.TooltipNumber:
                element_class = TooltipNumber
            case ComponentType.UpDownCounter:
                element_class = UpDownCounter
            case ComponentType.VideoSwitchbox:
                element_class = VideoSwitchbox
            case ComponentType.XOR:
                element_class = XOR
        if not element_class:
            raise NotImplementedError(
                f"No Component found for component type {component_type}"
            )
        return element_class.from_xml(element)


from .Components import *  # noqa: ignore=F403 pylint: disable=wildcard-import,wrong-import-position



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class ArithmeticFunction1In(Component):
    """
    Evaluates a mathematical expression with 1 input variable and outputs the result.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        function: str,
        x_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.ArithmeticFunction1In, component_id, position, 1.0
        )
        self.function: str = function
        self.x_input: Optional[Input] = x_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> ArithmeticFunction1In:
        assert element.tag == "c", f"invalid ArithmeticFunction1In {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.ArithmeticFunction1In.value
        ), f"Not an ArithmeticFunction1In {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = ArithmeticFunction1In._basic_in_parsing(obj)
        function: str = obj.attributes.get("e", "")
        return ArithmeticFunction1In(
            component_id,
            position,
            function,
            inputs.get("1"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {"e": self.function}
        children: list[XMLParserElement] = self._pos_in_to_xml(self.x_input)
        return attributes, children



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PropertyToggle(Component):
    """
    Adds a custom on/off toggle that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        name: str = "toggle",
        on_label: str = "on",
        off_label: str = "off",
        value_property: bool = False,
    ):
        super().__init__(ComponentType.PropertyToggle, component_id, position, 0.5)
        self.name: str = name
        self.on_label: str = on_label
        self.off_label: str = off_label
        self.value_property: bool = value_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertyToggle:
        assert element.tag == "c", f"invalid PropertyToggle {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertyToggle.value
        ), f"Not an PropertyToggle {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, _ = PropertyToggle._basic_in_parsing(obj)
        name: str = obj.attributes.get("n", "toggle")
        on_label: str = obj.attributes.get("on", "on")
        off_label: str = obj.attributes.get("off", "off")
        value_property: bool = obj.attributes.get("v", "false") == "true"
        return PropertyToggle(
            component_id, position, name, on_label, off_label, value_property
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {
            "n": self.name,
            "on": self.on_label,
            "off": self.off_label,
            "v": str(self.value_property).lower(),
        }
        children: list[XMLParserElement] = self._pos_in_to_xml()
        return attributes, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

from .SubTypes.MinMaxComponent import MinMaxComponent
from .SubTypes.ResetComponent import ResetComponent


class UpDownCounter(MinMaxComponent, ResetComponent):
    """
    Has an internal value that will increase and decrease when receiving different signals.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        up_input: Optional[Input] = None,
        down_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
        min_property: Optional[NumberProperty] = None,
        max_property: Optional[NumberProperty] = None,
        reset_property: Optional[NumberProperty] = None,
        increment_property: Optional[NumberProperty] = None,
    ):
        super().__init__(
            ComponentType.UpDownCounter,
            component_id,
            position,
            1.0,
            min_property,
            max_property,
            reset_property=reset_property,
        )
        self.up_input: Optional[Input] = up_input
        self.down_input: Optional[Input] = down_input
        self.reset_input: Optional[Input] = reset_input
        self.increment_property: NumberProperty = increment_property or NumberProperty(
            "0", "i"
        )

    @staticmethod
    def from_xml(element: XMLParserElement) -> UpDownCounter:
        assert element.tag == "c", f"invalid UpDownCounter {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.UpDownCounter.value
        ), f"Not an UpDownCounter {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = UpDownCounter._basic_in_parsing(
            obj
        )
        min_property, max_property = UpDownCounter._basic_min_max_parsing(properties)
        reset_property = UpDownCounter._basic_reset_parsing(properties)
        increment_property: Optional[NumberProperty] = properties.get("i")
        return UpDownCounter(
            component_id,
            position,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            min_property,
            max_property,
            reset_property,
            increment_property,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.up_input, self.down_input, self.reset_input
        )
        children.extend(self._min_max_to_xml())
        children.extend(self._reset_to_xml())
        children.append(self.increment_property.to_xml())
        return {}, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PIDController(Component):
    """
    A basic PID controller. The proportional, integral and derivative gains can be set in the property panel.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        setpoint_input: Optional[Input] = None,
        process_variable_input: Optional[Input] = None,
        active_input: Optional[Input] = None,
        proportional_property: Optional[NumberProperty] = None,
        integral_property: Optional[NumberProperty] = None,
        derivative_property: Optional[NumberProperty] = None,
    ):
        super().__init__(ComponentType.PIDController, component_id, position, 1.0)
        self.setpoint_input: Optional[Input] = setpoint_input
        self.process_variable_input: Optional[Input] = process_variable_input
        self.active_input: Optional[Input] = active_input
        self.proportional_property: NumberProperty = (
            proportional_property or NumberProperty("0", "kp")
        )
        self.integral_property: NumberProperty = integral_property or NumberProperty(
            "0", "ki"
        )
        self.derivative_property: NumberProperty = (
            derivative_property or NumberProperty("0", "kd")
        )

    @staticmethod
    def from_xml(element: XMLParserElement) -> PIDController:
        assert element.tag == "c", f"invalid PIDController {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PIDController.value
        ), f"Not an PIDController {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = PIDController._basic_in_parsing(
            obj
        )
        proportional_property: Optional[NumberProperty] = properties.get("kp")
        integral_property: Optional[NumberProperty] = properties.get("ki")
        derivative_property: Optional[NumberProperty] = properties.get("kd")
        return PIDController(
            component_id,
            position,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            proportional_property,
            integral_property,
            derivative_property,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.setpoint_input, self.process_variable_input, self.active_input
        )
        children.append(self.proportional_property.to_xml())
        children.append(self.integral_property.to_xml())
        children.append(self.derivative_property.to_xml())
        return {}, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class NumericalJunction(Component):
    """
    Outputs the input number to one of the outputs depending on whether or not the Switch Signal is on.
    The path that the input doesn't take will output a value of 0.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        value_to_pass_through_input: Optional[Input] = None,
        switch_signal_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.NumericalJunction, component_id, position, 0.75)
        self.value_to_pass_through_input: Optional[Input] = value_to_pass_through_input
        self.switch_signal_input: Optional[Input] = switch_signal_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> NumericalJunction:
        assert element.tag == "c", f"invalid NumericalJunction {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.NumericalJunction.value
        ), f"Not an NumericalJunction {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = NumericalJunction._basic_in_parsing(obj)
        return NumericalJunction(
            component_id, position, inputs.get("1"), inputs.get("2")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.value_to_pass_through_input, self.switch_signal_input
        )



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PropertyText(Component):
    """
    Adds a custom text input that will be seen on the microcontroller's property panel when placed on a vehicle,
    for showing player-defined text within a lua script.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        name: str = "Label",
        value_property: str = "",
    ):
        super().__init__(ComponentType.PropertyText, component_id, position, 0.5)
        self.name: str = name
        self.value_property: str = value_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertyText:
        assert element.tag == "c", f"invalid PropertyText {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertyText.value
        ), f"Not an PropertyText {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, _ = PropertyText._basic_in_parsing(obj)
        name: str = obj.attributes.get("n", "toggle")
        value_property: str = obj.attributes.get("v", "")
        return PropertyText(component_id, position, name, value_property)

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {
            "n": self.name,
            "v": self.value_property,
        }
        children: list[XMLParserElement] = self._pos_in_to_xml()
        return attributes, children



import unittest

from sw_mc_lib.XMLParser import XMLParser, XMLParserElement, parse


class TestXMLParser(unittest.TestCase):
    def test_unfinished_element(self) -> None:
        with self.assertRaises(NameError):
            parse("<abc ")
        with self.assertRaises(NameError):
            parse("<abc")

    def test_unfinished_string(self) -> None:
        with self.assertRaises(NameError):
            parse('<abc def="ghi')

    def test_unclosed_element(self) -> None:
        with self.assertRaises(NameError):
            parse("<abc>def")

    def test_wrong_closed_element(self) -> None:
        with self.assertRaises(NameError):
            parse("<abc></def>")

    def test_empty_element(self) -> None:
        res: XMLParserElement = parse("<abc/>")
        self.assertEqual(res, XMLParserElement("abc"))
        self.assertNotEqual(res, XMLParserElement("def"))
        self.assertNotEqual(res, "abc")

    def test_element_with_attributes(self) -> None:
        res: XMLParserElement = parse('<abc def="ghi" jkl="mno"/>')
        self.assertEqual(res, XMLParserElement("abc", {"def": "ghi", "jkl": "mno"}))

    def test_element_with_children(self) -> None:
        res: XMLParserElement = parse("<abc><def><ghi/></def></abc>")
        self.assertEqual(
            res,
            XMLParserElement(
                "abc", {}, [XMLParserElement("def", {}, [XMLParserElement("ghi")])]
            ),
        )

    def test_line_numbers(self) -> None:
        parser = XMLParser("<abc/>\n")
        parser.parse()
        self.assertEqual(parser.line, 2)
        self.assertEqual(parser.column, 1)

    def test_skipping_header(self) -> None:
        res: XMLParserElement = parse("<? ... ?><abc/>")
        self.assertEqual(res, XMLParserElement("abc"))
        with self.assertRaises(NameError):
            parse("<? abc")



import unittest

from sw_mc_lib.Input import Input, XMLParserElement


class TestInput(unittest.TestCase):
    def test_from_xml(self) -> None:
        elem: XMLParserElement = XMLParserElement("in1")
        expected: Input = Input(0, 0, "1")
        self.assertEqual(Input.from_xml(elem), expected)
        elem.attributes["component_id"] = "1234"
        expected.component_id = 1234
        self.assertEqual(Input.from_xml(elem), expected)
        elem.attributes["node_index"] = "2"
        expected.node_index = 2
        self.assertEqual(Input.from_xml(elem), expected)
        elem.tag = "inoff"
        expected.index = "off"
        self.assertEqual(Input.from_xml(elem), expected)

    def test_to_xml(self) -> None:
        current: Input = Input(123)
        expected: XMLParserElement = XMLParserElement("intemp", {"component_id": "123"})
        self.assertEqual(current.to_xml(), expected)
        current.index = "1"
        expected.tag = "in1"
        self.assertEqual(current.to_xml(), expected)
        current.node_index = 3
        expected.attributes["node_index"] = "3"
        self.assertEqual(current.to_xml(), expected)



from enum import Enum


class _Enum(Enum):
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"


class SignalType(_Enum):
    """
    Signal types
    """

    Boolean: int = 0
    Number: int = 1
    Composite: int = 5
    Video: int = 6
    Audio: int = 7


class NodeMode(_Enum):
    """
    Modes of nodes (in/out connections)
    """

    Output: int = 0
    Input: int = 1


class TooltipMode(_Enum):
    """
    Modes of Tooltips, depending on the error/signal input
    """

    Always: int = 0
    IfOn: int = 1
    IfOff: int = 2


class PulseMode(_Enum):
    """
    Modes of the Pulse Component, note that OffToOn is the default mode
    """

    OnToOff: int = 0
    OffToOn: int = 1
    Always: int = 2


class TimerUnit(_Enum):
    """
    Units of timer Components
    """

    Seconds: int = 0
    Ticks: int = 1


class ComponentType(_Enum):
    """
    Types of Components and their respective ids
    """

    NOT: int = 0
    AND: int = 1
    OR: int = 2
    XOR: int = 3
    NAND: int = 4
    NOR: int = 5
    Add: int = 6
    Subtract: int = 7
    Multiply: int = 8
    Divide: int = 9
    ArithmeticFunction3In: int = 10
    Clamp: int = 11
    Threshold: int = 12
    MemoryRegister: int = 13
    Abs: int = 14
    ConstantNumber: int = 15
    ConstantOn: int = 16
    GreaterThan: int = 17
    LessThan: int = 18
    PropertySlider: int = 19
    PropertyDropdown: int = 20
    NumericalJunction: int = 21
    NumericalSwitchbox: int = 22
    PIDController: int = 23
    SRLatch: int = 24
    JKFlipFlop: int = 25
    Capacitor: int = 26
    Blinker: int = 27
    PushToToggle: int = 28
    CompositeReadBoolean: int = 29
    CompositeWriteBooleanDeprecated: int = 30
    CompositeReadNumber: int = 31
    CompositeWriteNumberDeprecated: int = 32
    PropertyToggle: int = 33
    PropertyNumber: int = 34
    Delta: int = 35
    ArithmeticFunction8In: int = 36
    UpDownCounter: int = 37
    Modulo: int = 38
    PIDControllerAdvanced: int = 39
    CompositeWriteNumber: int = 40
    CompositeWriteBoolean: int = 41
    Equal: int = 42
    TooltipNumber: int = 43
    TooltipBoolean: int = 44
    ArithmeticFunction1In: int = 45
    BooleanFunction4In: int = 46
    BooleanFunction8In: int = 47
    Pulse: int = 48
    TimerTON: int = 49
    TimerTOF: int = 50
    TimerRTO: int = 51
    TimerRTF: int = 52
    CompositeSwitchbox: int = 53
    NumberToCompositeBinary: int = 54
    CompositeBinaryToNumber: int = 55
    LuaScript: int = 56
    VideoSwitchbox: int = 57
    PropertyText: int = 58
    AudioSwitchbox: int = 59



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class CompositeBinaryToNumber(Component):
    """
    Reads the on/off signals of a composite link and encodes them in the bits of an output number.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        signal_to_convert_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.CompositeBinaryToNumber, component_id, position, 0.5
        )
        self.signal_to_convert_input: Optional[Input] = signal_to_convert_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> CompositeBinaryToNumber:
        assert element.tag == "c", f"invalid CompositeBinaryToNumber {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.CompositeBinaryToNumber.value
        ), f"Not an CompositeBinaryToNumber {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = CompositeBinaryToNumber._basic_in_parsing(
            obj
        )
        return CompositeBinaryToNumber(component_id, position, inputs.get("1"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(self.signal_to_convert_input)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PIDControllerAdvanced(Component):
    """
    A PID controller with variable proportional, integral and derivative gains.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        setpoint_input: Optional[Input] = None,
        process_variable_input: Optional[Input] = None,
        proportional_input: Optional[Input] = None,
        integral_input: Optional[Input] = None,
        derivative_input: Optional[Input] = None,
        active_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.PIDControllerAdvanced, component_id, position, 2.25
        )
        self.setpoint_input: Optional[Input] = setpoint_input
        self.process_variable_input: Optional[Input] = process_variable_input
        self.proportional_input: Optional[Input] = proportional_input
        self.integral_input: Optional[Input] = integral_input
        self.derivative_input: Optional[Input] = derivative_input
        self.active_input: Optional[Input] = active_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> PIDControllerAdvanced:
        assert element.tag == "c", f"invalid PIDControllerAdvanced {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PIDControllerAdvanced.value
        ), f"Not an PIDControllerAdvanced {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = PIDControllerAdvanced._basic_in_parsing(obj)
        return PIDControllerAdvanced(
            component_id,
            position,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            inputs.get("4"),
            inputs.get("5"),
            inputs.get("6"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.setpoint_input,
            self.process_variable_input,
            self.proportional_input,
            self.integral_input,
            self.derivative_input,
            self.active_input,
        )



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class AudioSwitchbox(Component):
    """
    Outputs the first input audio when receiving an on signal, and the second when receiving an off signal.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        on_value_input: Optional[Input] = None,
        off_value_input: Optional[Input] = None,
        switch_signal_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.AudioSwitchbox, component_id, position, 1.0)
        self.on_value_input: Optional[Input] = on_value_input
        self.off_value_input: Optional[Input] = off_value_input
        self.switch_signal_input: Optional[Input] = switch_signal_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> AudioSwitchbox:
        assert element.tag == "c", f"invalid AudioSwitchbox {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.AudioSwitchbox.value
        ), f"Not an AudioSwitchbox {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = AudioSwitchbox._basic_in_parsing(obj)
        return AudioSwitchbox(
            component_id, position, inputs.get("1"), inputs.get("2"), inputs.get("3")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(
            self.on_value_input, self.off_value_input, self.switch_signal_input
        )



from __future__ import annotations

from typing import Optional

from .XMLElement import XMLElement, XMLParserElement


class Input(XMLElement):
    """
    An input for a Component. Contains both component_id as well as node_index of the source.
    """

    def __init__(
        self, component_id: int, node_index: int = 0, index: Optional[str] = None
    ):
        self.index: str = index or "temp"
        self.component_id: int = component_id
        self.node_index: int = node_index

    @staticmethod
    def from_xml(element: XMLParserElement) -> Input:
        assert element.tag.startswith("in"), f"invalid Input {element}"
        index: str = element.tag.replace("in", "", 1)
        component_id: int = int(element.attributes.get("component_id", "0"))
        node_index: int = int(element.attributes.get("node_index", "0"))
        return Input(component_id, node_index, index)

    def to_xml(self) -> XMLParserElement:
        attributes: dict[str, str] = {"component_id": str(self.component_id)}
        if self.node_index != 0:
            attributes["node_index"] = str(self.node_index)
        return XMLParserElement(f"in{self.index}", attributes)



from __future__ import annotations

from typing import Optional

import numpy as np

from .XMLElement import XMLElement, XMLParserElement


def string_to_sw_float(to_convert: str) -> float:
    """
    Converts a number string to a float like the game does right now.

    :param to_convert: The string containing a float number
    :return: The resulting float
    """
    return float(f"{np.float32(to_convert):0.6f}")


class NumberProperty(XMLElement):
    """
    A property with a numerical value. Stores the number as string to preserve full precision and
    leading/trailing zeroes. Name is how it is referred to by the game, and will be filled in upon serializing.
    """

    def __init__(self, text: str, name: Optional[str] = None):
        self.text: str = text
        self.name: str = name or "temp"

    @staticmethod
    def from_xml(element: XMLParserElement) -> NumberProperty:
        name: str = element.tag
        text: str = element.attributes.get("text", "0")
        return NumberProperty(text, name)

    def to_xml(self) -> XMLParserElement:
        return XMLParserElement(
            self.name, {"text": self.text, "value": str(self.value)}
        )

    @property
    def value(self) -> float:
        """
        The value of the text field as float property.

        :return: Float of text
        """
        return string_to_sw_float(self.text)

    @value.setter
    def value(self, value: float) -> None:
        self.text = str(value)



from __future__ import annotations

import string
from typing import Any, Optional


class XMLParserElement:
    """
    An element that is similar to XML elements.
    """

    def __init__(
        self,
        tag: str,
        attributes: Optional[dict[str, str]] = None,
        children: Optional[list[XMLParserElement]] = None,
    ):
        self.tag: str = tag
        self.attributes: dict[str, str] = attributes or {}
        self.children: list[XMLParserElement] = children or []

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, XMLParserElement):
            return (
                self.tag == other.tag
                and self.attributes == other.attributes
                and self.children == other.children
            )
        return False

    def __repr__(self) -> str:
        return f"XMLParserElement(tag={self.tag!r}, attributes={self.attributes!r}, children={self.children!r}"


class XMLParser:
    """
    Parser to parse strings into XML elements. To parse XML, you probably want to use
    :func:`sw_mc_lib.XMLParser.parse`.
    """

    IDENTIFIER_CHARACTERS: str = string.ascii_letters + string.digits + "_"

    def __init__(self, content: str):
        self.content: str = content
        self.pos: int = 0
        self.line: int = 1
        self.column: int = 1
        self.current: Optional[str] = content[0]

    def advance(self) -> None:
        """
        Advance the pointer by one and update line, column and current

        :return: None
        """
        self.pos += 1
        if self.pos >= len(self.content):
            self.current = None
            return
        self.current = self.content[self.pos]
        if self.current == "\n":
            self.line += 1
            self.column = 0
        self.column += 1

    def skip_whitespace(self) -> None:
        """
        Skip all kinds of whitespace characters

        :return: None
        """
        while self.current and self.current.isspace():
            self.advance()

    def read_name(self) -> str:
        """
        Read a name (so a string consisting of IDENTIFIER_CHARACTERS until current is not such a character

        :return: name
        """
        name: str = ""
        while self.current and self.current in self.IDENTIFIER_CHARACTERS:
            name += self.current
            self.advance()
        return name

    def read_and_unescape_string(self) -> str:
        """
        Read a n attribute string, so a string constrained either by `'` or `"`

        :return: The unescaped resulting string
        """
        res_str: str = ""
        assert self.current
        opening_char: str = self.current
        self.advance()
        while self.current and self.current != opening_char:
            res_str += self.current
            self.advance()
        self.advance()
        res_str = res_str.replace("&lt;", "<")
        res_str = res_str.replace("&gt;", ">")
        res_str = res_str.replace("&apos;", "'")
        res_str = res_str.replace("&quot;", '"')
        res_str = res_str.replace("&amp;", "&")
        return res_str

    def error(self, expected: str, received: str, line: int, column: int) -> None:
        """
        Raise an error

        :param expected: The character or string that has been expected at the position
        :param received: The character or string that has been received instead
        :param line: The line it occurred on
        :param column: The column it occurred on
        :return: None
        """
        raise NameError(f"Expected {expected} at {line}:{column}, got {received}")

    def eat(self, expected: str) -> None:
        """
        Eat a character, throwing an error if it is unlike the expected character

        :param expected: The expected character
        :return: None
        """
        if self.current == expected:
            self.advance()
        else:
            self.error(expected, self.current or "<EOL>", self.line, self.column)

    def read_element(self) -> XMLParserElement:
        """
        Read an XML element, with a tag, attributes and children

        :return: XML element
        """
        tag: str = self.read_name()
        self.skip_whitespace()
        attributes: dict[str, str] = {}
        while self.current and self.current in self.IDENTIFIER_CHARACTERS:
            attr_name: str = self.read_name()
            self.eat("=")
            attr_value: str = self.read_and_unescape_string()
            attributes[attr_name] = attr_value
            self.skip_whitespace()
        children: list[XMLParserElement] = []
        if self.current == ">":
            self.eat(">")
            self.skip_whitespace()
            self.eat("<")
            while self.current and self.current != "/":
                children.append(self.read_element())
                self.skip_whitespace()
                self.eat("<")
            self.advance()
            line, column = self.line, self.column
            closing_tag: str = self.read_name()
            if closing_tag != tag:
                self.error(tag, closing_tag, line, column)
            self.eat(">")
        else:
            self.eat("/")
            self.eat(">")
        return XMLParserElement(tag, attributes, children)

    def parse(self) -> XMLParserElement:
        """
        Parse the string, discarding the XML declaration

        :return: The parsed element
        """
        self.eat("<")
        if self.current == "?":
            self.advance()
            while self.current and self.current != "?":
                self.advance()
            self.eat("?")
            self.eat(">")
            self.skip_whitespace()
            self.eat("<")
        return self.read_element()


def parse(content: str) -> XMLParserElement:
    """
    Parse a string to XML elements

    :param content: The string to parse
    :return: The root XML element
    """
    return XMLParser(content).parse()



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

DEFAULT_SCRIPT: str = """-- Tick function that will be executed every logic tick
function onTick()
\tvalue = input.getNumber(1)\t\t\t -- Read the first number from the script's composite input
\toutput.setNumber(1, value * 10)\t\t-- Write a number to the script's composite output
end

-- Draw function that will be executed when this script renders to a screen
function onDraw()
\tw = screen.getWidth()\t\t\t\t  -- Get the screen's width and height
\th = screen.getHeight()\t\t\t\t\t
\tscreen.setColor(0, 255, 0)\t\t\t -- Set draw color to green
\tscreen.drawCircleF(w / 2, h / 2, 30)   -- Draw a 30px radius circle in the center of the screen
end"""


class LuaScript(Component):
    """
    Runs a custom lua script for advanced logic and drawing to monitors.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        script_property: str = DEFAULT_SCRIPT,
        data_input: Optional[Input] = None,
        video_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.LuaScript, component_id, position, 0.75)
        self.data_input: Optional[Input] = data_input
        self.video_input: Optional[Input] = video_input
        self.script_property: str = script_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> LuaScript:
        assert element.tag == "c", f"invalid LuaScript {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.LuaScript.value
        ), f"Not an LuaScript {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = LuaScript._basic_in_parsing(obj)
        script_property: str = obj.attributes.get("script", DEFAULT_SCRIPT)
        return LuaScript(
            component_id, position, script_property, inputs.get("1"), inputs.get("2")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"script": self.script_property}, self._pos_in_to_xml(
            self.data_input, self.video_input
        )



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

from .SubTypes.ResetComponent import ResetComponent


class MemoryRegister(ResetComponent):
    """
    "Remembers the input value when receiving a signal to the Set node. When the Reset node receives a signal,
    the stored number is cleared to a value that can be customised in the properties panel.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        reset_property: Optional[NumberProperty] = None,
        set_input: Optional[Input] = None,
        reset_input: Optional[Input] = None,
        number_to_store_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.MemoryRegister, component_id, position, 1.0, reset_property
        )
        self.set_input: Optional[Input] = set_input
        self.reset_input: Optional[Input] = reset_input
        self.number_to_store_input: Optional[Input] = number_to_store_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> MemoryRegister:
        assert element.tag == "c", f"invalid MemoryRegister {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.MemoryRegister.value
        ), f"Not an MemoryRegister {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = MemoryRegister._basic_in_parsing(
            obj
        )
        reset_property = MemoryRegister._basic_reset_parsing(properties)
        return MemoryRegister(
            component_id,
            position,
            reset_property,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.set_input, self.reset_input, self.number_to_store_input
        )
        children.extend(self._reset_to_xml())
        return {}, children



from __future__ import annotations

from abc import ABC
from typing import Optional

from sw_mc_lib.Component import Component
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class ResetComponent(Component, ABC):
    """
    A Component that has a reset value property.
    """

    def __init__(
        self,
        component_type: ComponentType,
        component_id: int,
        position: Position,
        height: float,
        reset_property: Optional[NumberProperty],
        **kwargs: Optional[NumberProperty],
    ):
        super().__init__(component_type, component_id, position, height, **kwargs)
        self.reset_property: NumberProperty = reset_property or NumberProperty("0", "r")

    @staticmethod
    def _basic_reset_parsing(
        properties: dict[str, NumberProperty]
    ) -> Optional[NumberProperty]:
        return properties.get("r")

    def _reset_to_xml(self) -> list[XMLParserElement]:
        return [self.reset_property.to_xml()]



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType, PulseMode
from sw_mc_lib.XMLParser import XMLParserElement


class Pulse(Component):
    """
    A switch that outputs a single tick pulse.
    It can be configured to pulse when being switched from off to on (default), on to off,
    or always when the input signal changes.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        mode_property: PulseMode = PulseMode.OffToOn,
        toggle_signal_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.Pulse, component_id, position, 0.5)
        self.toggle_signal_input: Optional[Input] = toggle_signal_input
        self.mode_property: PulseMode = mode_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> Pulse:
        assert element.tag == "c", f"invalid Pulse {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.Pulse.value
        ), f"Not an Pulse {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Pulse._basic_in_parsing(obj)
        mode_property: PulseMode = PulseMode(int(obj.attributes.get("m", "1")))
        return Pulse(component_id, position, mode_property, inputs.get("1"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {"m": str(self.mode_property.value)}, self._pos_in_to_xml(
            self.toggle_signal_input
        )



import unittest

from sw_mc_lib.XMLFormatter import format
from sw_mc_lib.XMLParser import XMLParserElement


class TestXMLFormatter(unittest.TestCase):
    def test_empty_element_formatting(self) -> None:
        res: str = format(XMLParserElement("abc"))
        self.assertEqual(res, "<abc/>\n")

    def test_element_with_attributes(self) -> None:
        res: str = format(XMLParserElement("abc", {"def": "ghi"}))
        self.assertEqual(res, '<abc def="ghi"/>\n')

    def test_element_with_children(self) -> None:
        res: str = format(
            XMLParserElement(
                "abc",
                {"def": "ghi"},
                [XMLParserElement("jkl"), XMLParserElement("mno")],
            )
        )
        self.assertEqual(res, '<abc def="ghi">\n\t<jkl/>\n\t<mno/>\n</abc>\n')

    def test_element_with_children_with_attributes(self) -> None:
        res: str = format(
            XMLParserElement(
                "abc",
                {},
                [XMLParserElement("def", {"ghi": "jkl"}, [XMLParserElement("mno")])],
            )
        )
        self.assertEqual(
            res, '<abc>\n\t<def ghi="jkl">\n\t\t<mno/>\n\t</def>\n</abc>\n'
        )

    def test_four_spaces_indent(self) -> None:
        res: str = format(
            XMLParserElement(
                "abc", {}, [XMLParserElement("def", {}, [XMLParserElement("ghi")])]
            ),
            indentation="    ",
        )
        self.assertEqual(res, "<abc>\n    <def>\n        <ghi/>\n    </def>\n</abc>\n")

    def test_empty_indent(self) -> None:
        res: str = format(
            XMLParserElement(
                "abc", {}, [XMLParserElement("def", {}, [XMLParserElement("ghi")])]
            ),
            indentation="",
        )
        self.assertEqual(res, "<abc>\n<def>\n<ghi/>\n</def>\n</abc>\n")

    def test_no_indent(self) -> None:
        res: str = format(
            XMLParserElement(
                "abc", {}, [XMLParserElement("def", {}, [XMLParserElement("ghi")])]
            ),
            indentation=None,
        )
        self.assertEqual(res, "<abc><def><ghi/></def></abc>")
        res = format(XMLParserElement("abc", {"def": "a\nbc"}), indentation=None)
        self.assertEqual(res, '<abc def="a\nbc"/>')

    def test_string_escape(self) -> None:
        res: str = format(XMLParserElement("abc", {"def": 'g\nh<>&"i'}))
        self.assertEqual(res, "<abc def='g\nh&lt;&gt;&amp;\"i'/>\n")
        res = format(XMLParserElement("abc", {"def": 'g\'h""i'}))
        self.assertEqual(res, "<abc def='g&apos;h\"\"i'/>\n")
        res = format(XMLParserElement("abc", {"def": "g\"h''i"}))
        self.assertEqual(res, "<abc def=\"g&quot;h''i\"/>\n")

    def test_attributes_with_newlines_indent(self) -> None:
        res: str = format(
            XMLParserElement("abc", {}, [XMLParserElement("def", {"def": "ghi\njkl"})])
        )
        self.assertEqual(res, '<abc>\n\t<def def="ghi\njkl"/>\n</abc>\n')



import unittest

from sw_mc_lib.Position import Position, XMLParserElement


class TestPosition(unittest.TestCase):
    def test_from_xml(self) -> None:
        elem: XMLParserElement = XMLParserElement("pos")
        expected: Position = Position(0, 0)
        self.assertEqual(Position.from_xml(elem), expected)
        elem.attributes["x"] = "1.25"
        expected.x = 1.25
        self.assertEqual(Position.from_xml(elem), expected)
        elem.attributes["y"] = "-4.75"
        expected.y = -4.75
        self.assertEqual(Position.from_xml(elem), expected)

    def test_empty_pos(self) -> None:
        self.assertEqual(Position(), Position(0, 0))

    def test_to_xml(self) -> None:
        elem: Position = Position(0, 0)
        expected: XMLParserElement = XMLParserElement("pos")
        self.assertEqual(elem.to_xml(), expected)
        elem.x = 1.25
        expected.attributes["x"] = "1.25"
        self.assertEqual(elem.to_xml(), expected)
        elem.y = -4.0
        expected.attributes["y"] = "-4.0"
        self.assertEqual(elem.to_xml(), expected)



from __future__ import annotations

from abc import ABC
from typing import Optional

from sw_mc_lib.Component import Component
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class MinMaxComponent(Component, ABC):
    """
    A Component that has a min and max Property.
    """

    def __init__(
        self,
        component_type: ComponentType,
        component_id: int,
        position: Position,
        height: float,
        min_property: Optional[NumberProperty] = None,
        max_property: Optional[NumberProperty] = None,
        **kwargs: Optional[NumberProperty],
    ):
        super().__init__(component_type, component_id, position, height, **kwargs)
        self.min_property: NumberProperty = min_property or NumberProperty("0", "min")
        self.max_property: NumberProperty = max_property or NumberProperty("0", "max")

    @staticmethod
    def _basic_min_max_parsing(
        properties: dict[str, NumberProperty]
    ) -> tuple[Optional[NumberProperty], Optional[NumberProperty]]:
        return properties.get("min"), properties.get("max")

    def _min_max_to_xml(self) -> list[XMLParserElement]:
        return [self.min_property.to_xml(), self.max_property.to_xml()]



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class PushToToggle(Component):
    """
    An on/off switch that is toggled every time a new on signal is sent to its input.
    """

    def __init__(
        self, component_id: int, position: Position, a_input: Optional[Input] = None
    ):
        super().__init__(ComponentType.PushToToggle, component_id, position, 0.5)
        self.a_input: Optional[Input] = a_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> PushToToggle:
        assert element.tag == "c", f"invalid PushToToggle {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PushToToggle.value
        ), f"Not an PushToToggle {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = PushToToggle._basic_in_parsing(obj)
        return PushToToggle(component_id, position, inputs.get("1"))

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(self.a_input)



from __future__ import annotations

import unittest
from typing import Optional

from sw_mc_lib.Component import Component, ComponentType
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.XMLParser import XMLParserElement


class TestingComponent(Component):
    def __init__(
        self,
        component_id: int,
        position: Position,
        attributes: dict[str, str],
        inputs: dict[str, Input],
        properties: dict[str, NumberProperty],
    ):
        super().__init__(ComponentType.Abs, component_id, position, 1.0)
        self.attributes: dict[str, str] = attributes
        self.inputs: dict[str, Input] = inputs
        self.properties: dict[str, NumberProperty] = properties

    @staticmethod
    def from_xml(element: XMLParserElement) -> TestingComponent:
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = TestingComponent._basic_in_parsing(
            obj
        )
        attributes: dict[str, str] = {
            name: value for name, value in obj.attributes.items() if name != "id"
        }
        return TestingComponent(component_id, position, attributes, inputs, properties)

    def _inner_to_xml(self) -> tuple[dict[str, str], list[XMLParserElement]]:
        input_list: list[Optional[Input]] = []
        named_inputs: dict[str, Optional[Input]] = {}
        for name, input in self.inputs.items():
            if name.isdigit():
                while len(input_list) < int(name):
                    input_list.append(None)
                input_list[int(name) - 1] = input
            else:
                named_inputs[name] = input
        return self.attributes, self._pos_in_to_xml(
            *input_list,
            named_inputs=named_inputs or None,
            properties=self.properties or None,
        )


class TestComponent(unittest.TestCase):
    def test_basic_in_parsing(self) -> None:
        elem: XMLParserElement = XMLParserElement("object", {"id": "123"})
        expected: tuple[int, Position, dict[str, Input], dict[str, NumberProperty]] = (
            123,
            Position(),
            {},
            {},
        )
        self.assertEqual(Component._basic_in_parsing(elem), expected)
        elem.children.append(XMLParserElement("in1", {"component_id": "1"}))
        expected[2]["1"] = Input(1, 0, "1")
        self.assertEqual(Component._basic_in_parsing(elem), expected)
        elem.children.append(XMLParserElement("in3", {"component_id": "145"}))
        expected[2]["3"] = Input(145, 0, "3")
        self.assertEqual(Component._basic_in_parsing(elem), expected)
        elem.children.append(XMLParserElement("property"))
        expected[3]["property"] = NumberProperty("0", "property")
        self.assertEqual(Component._basic_in_parsing(elem), expected)
        expected[3]["prop"] = NumberProperty("5", "prop")
        elem.children.append(expected[3]["prop"].to_xml())
        self.assertEqual(Component._basic_in_parsing(elem), expected)

    def test_pos_in_to_xml(self) -> None:
        pos: Position = Position(10, 5)
        test_comp: TestingComponent = TestingComponent(100, pos, {}, {}, {})
        expected: list[XMLParserElement] = [pos.to_xml()]
        _, children = test_comp._inner_to_xml()
        self.assertEqual(children, expected)
        test_comp.inputs["1"] = Input(10)
        expected.append(Input(10, 0, "1").to_xml())
        _, children = test_comp._inner_to_xml()
        self.assertEqual(children, expected)
        test_comp.inputs["5"] = Input(500, 2)
        expected.append(Input(500, 2, "5").to_xml())
        _, children = test_comp._inner_to_xml()
        self.assertEqual(children, expected)
        test_comp.inputs["abc"] = Input(30)
        expected.append(Input(30, 0, "abc").to_xml())
        _, children = test_comp._inner_to_xml()
        self.assertEqual(children, expected)
        test_comp.properties["abc"] = NumberProperty("3")
        expected.append(NumberProperty("3", "abc").to_xml())
        _, children = test_comp._inner_to_xml()
        self.assertEqual(children, expected)
        test_comp.attributes["abc"] = "def"
        attributes, _ = test_comp._inner_to_xml()
        self.assertEqual(attributes, {"abc": "def"})

    def test_to_xml(self) -> None:
        pos: Position = Position(4, 6)
        test_comp: TestingComponent = TestingComponent(40, pos, {}, {}, {})
        inner_object: XMLParserElement = XMLParserElement(
            "object", {"id": "40"}, [pos.to_xml()]
        )
        elem: XMLParserElement = XMLParserElement(
            "c", {"type": str(ComponentType.Abs.value)}, [inner_object]
        )
        self.assertEqual(test_comp.to_xml(), elem)
        test_comp.attributes["abc"] = "def"
        inner_object.attributes["abc"] = "def"
        self.assertEqual(test_comp.to_xml(), elem)



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType, TimerUnit
from sw_mc_lib.XMLParser import XMLParserElement


class TimerTON(Component):
    """
    Variable input timer. Outputs an on signal when the timer reaches its duration. The timer will reset when off.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        unit_property: TimerUnit = TimerUnit.Seconds,
        timer_enable_input: Optional[Input] = None,
        duration_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.TimerTON, component_id, position, 0.75)
        self.timer_enable_input: Optional[Input] = timer_enable_input
        self.duration_input: Optional[Input] = duration_input
        self.unit_property: TimerUnit = unit_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> TimerTON:
        assert element.tag == "c", f"invalid TimerTON {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.TimerTON.value
        ), f"Not an TimerTON {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = TimerTON._basic_in_parsing(obj)
        unit_property: TimerUnit = TimerUnit(int(obj.attributes.get("u", "0")))
        return TimerTON(
            component_id, position, unit_property, inputs.get("1"), inputs.get("2")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml(self.timer_enable_input, self.duration_input)



from __future__ import annotations

from itertools import chain

from .Component import Component
from .Node import Node
from .XMLElement import XMLElement, XMLParserElement


class Microcontroller(XMLElement):
    """
    A Stormworks Microcontroller
    """

    def __init__(
        self,
        name: str,
        description: str,
        width: int,
        length: int,
        nodes: list[Node],
        components: list[Component],
    ):
        self.name: str = name
        self.description: str = description
        self.width: int = width
        self.length: int = length
        self.nodes: list[Node] = nodes
        self.components: list[Component] = components

    @staticmethod
    def from_xml(element: XMLParserElement) -> Microcontroller:
        assert element.tag in ("microprocessor", "microprocessor_definition")
        name: str = element.attributes.get("name", "")
        description: str = element.attributes.get("description", "")
        width: int = int(element.attributes.get("width", "0"))
        length: int = int(element.attributes.get("length", "0"))
        nodes: list[Node] = []
        nodes_elem: XMLParserElement = element.children[0]
        assert nodes_elem.tag == "nodes"
        for node in nodes_elem.children:
            nodes.append(Node.from_xml(node))
        group_elem: XMLParserElement = element.children[1]
        assert group_elem.tag == "group"
        components_elem: XMLParserElement = group_elem.children[1]
        assert components_elem.tag == "components"
        components: list[Component] = []
        for component in components_elem.children:
            components.append(Component.from_xml(component))
        return Microcontroller(name, description, width, length, nodes, components)

    def to_xml(self) -> XMLParserElement:
        nodes_elem: XMLParserElement = XMLParserElement("nodes")
        for node in self.nodes:
            nodes_elem.children.append(node.to_xml())
        group_elem: XMLParserElement = XMLParserElement("group")
        group_elem.children.append(
            XMLParserElement(
                "data", {}, [XMLParserElement("inputs"), XMLParserElement("outputs")]
            )
        )
        components_elem: XMLParserElement = XMLParserElement("components")
        for component in self.components:
            components_elem.children.append(component.to_xml())
        group_elem.children.append(components_elem)
        group_elem.children.append(XMLParserElement("components_bridge"))
        group_elem.children.append(XMLParserElement("groups"))
        components_states_elem: XMLParserElement = XMLParserElement("component_states")
        for i, component in enumerate(self.components):
            components_states_elem.children.append(component.to_state_xml(i))
        group_elem.children.append(components_states_elem)
        group_elem.children.append(XMLParserElement("component_bridge_states"))
        group_elem.children.append(XMLParserElement("group_states"))
        attributes: dict[str, str] = {
            "name": self.name,
            "description": self.description,
            "width": str(self.width),
            "length": str(self.length),
            "id_counter": str(self.id_counter),
            "id_counter_node": str(self.node_counter),
        }
        return XMLParserElement("microprocessor", attributes, [nodes_elem, group_elem])

    @property
    def id_counter(self) -> int:
        """
        Stormworks internal counter of what the largest component_id is.

        :return: The currently largest component_id
        """
        components = (comp.component_id for comp in self.components)
        nodes = (node.component_id for node in self.nodes)
        return max(chain(components, nodes, (0,)))

    @property
    def node_counter(self) -> int:
        """
        Stormworks internal counter of what the largest node_id is.

        :return: The currently largest node_id
        """
        nodes = (node.node_id for node in self.nodes)
        return max(chain(nodes, (0,)))



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class Capacitor(Component):
    """
    Charges up when receiving an on signal, then discharges over a period of time.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        charge_input: Optional[Input] = None,
        charge_time_property: float = 1.0,
        discharge_time_property: float = 1.0,
    ):
        super().__init__(ComponentType.Capacitor, component_id, position, 1.0)
        self.charge_input: Optional[Input] = charge_input
        self.charge_time_property: float = charge_time_property
        self.discharge_time_property: float = discharge_time_property

    @staticmethod
    def from_xml(element: XMLParserElement) -> Capacitor:
        assert element.tag == "c", f"invalid Capacitor {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.Capacitor.value
        ), f"Not an Capacitor {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, _ = Capacitor._basic_in_parsing(obj)
        charge_time_property: float = float(obj.attributes.get("ct", "1"))
        discharge_time_property: float = float(obj.attributes.get("dt", "1"))
        return Capacitor(
            component_id,
            position,
            inputs.get("1"),
            charge_time_property,
            discharge_time_property,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {
            "ct": str(self.charge_time_property),
            "dt": str(self.discharge_time_property),
        }
        children: list[XMLParserElement] = self._pos_in_to_xml(self.charge_input)
        return attributes, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT
from sw_mc_lib.Input import Input
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

from .SubTypes.MinMaxComponent import MinMaxComponent


class Threshold(MinMaxComponent):
    """
    Outputs an on/off signal indicating whether or not the input value is within a set threshold.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        min_property: Optional[NumberProperty] = None,
        max_property: Optional[NumberProperty] = None,
        number_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.Threshold,
            component_id,
            position,
            0.5,
            min_property,
            max_property,
        )
        self.number_input: Optional[Input] = number_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> Threshold:
        assert element.tag == "c", f"invalid Threshold {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.Threshold.value
        ), f"Not an Threshold {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, inputs, properties = Threshold._basic_in_parsing(obj)
        min_property, max_property = Threshold._basic_min_max_parsing(properties)
        return Threshold(
            component_id, position, min_property, max_property, inputs.get("1")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml(self.number_input)
        children.extend(self._min_max_to_xml())
        return {}, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class CompositeReadNumber(Component):
    """
    Reads the number value from a selected channel of a composite input.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        channel_property: int = 1,
        composite_signal_input: Optional[Input] = None,
    ):
        super().__init__(
            ComponentType.CompositeReadNumber,
            component_id,
            position,
            0.5 if channel_property >= 1 else 0.75,
        )
        self.channel_property: int = channel_property
        self.composite_signal_input: Optional[Input] = composite_signal_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> CompositeReadNumber:
        assert element.tag == "c", f"invalid CompositeReadNumber {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.CompositeReadNumber.value
        ), f"Not an CompositeReadNumber {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = CompositeReadNumber._basic_in_parsing(obj)
        channel_property: int = int(obj.attributes.get("i", "0")) + 1
        return CompositeReadNumber(
            component_id, position, channel_property, inputs.get("1")
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {"i": str(self.channel_property - 1)}
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.composite_signal_input
        )
        return attributes, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement

from .SubTypes.MinMaxComponent import MinMaxComponent
from .SubTypes.ValueComponent import ValueComponent


class PropertySlider(MinMaxComponent, ValueComponent):
    """
    Adds a custom slider that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        min_property: Optional[NumberProperty] = None,
        max_property: Optional[NumberProperty] = None,
        value_property: Optional[NumberProperty] = None,
        rounding_property: Optional[NumberProperty] = None,
    ):
        super().__init__(
            ComponentType.PropertySlider,
            component_id,
            position,
            0.5,
            min_property,
            max_property,
            value_property=value_property,
        )
        self.rounding_property: NumberProperty = rounding_property or NumberProperty(
            "0", "int"
        )

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertySlider:
        assert element.tag == "c", f"invalid PropertySlider {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertySlider.value
        ), f"Not an PropertySlider {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, properties = PropertySlider._basic_in_parsing(obj)
        min_property, max_property = PropertySlider._basic_min_max_parsing(properties)
        value_property = PropertySlider._basic_value_parsing(properties)
        rounding_property: Optional[NumberProperty] = properties.get("int")
        return PropertySlider(
            component_id,
            position,
            min_property,
            max_property,
            value_property,
            rounding_property,
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        children: list[XMLParserElement] = self._pos_in_to_xml()
        children.extend(self._min_max_to_xml())
        children.extend(self._value_to_xml())
        children.append(self.rounding_property.to_xml())
        return {}, children



from .Abs import Abs
from .Add import Add
from .AND import AND
from .ArithmeticFunction1In import ArithmeticFunction1In
from .ArithmeticFunction3In import ArithmeticFunction3In
from .ArithmeticFunction8In import ArithmeticFunction8In
from .AudioSwitchbox import AudioSwitchbox
from .Blinker import Blinker
from .BooleanFunction4In import BooleanFunction4In
from .BooleanFunction8In import BooleanFunction8In
from .Capacitor import Capacitor
from .Clamp import Clamp
from .CompositeBinaryToNumber import CompositeBinaryToNumber
from .CompositeReadBoolean import CompositeReadBoolean
from .CompositeReadNumber import CompositeReadNumber
from .CompositeSwitchbox import CompositeSwitchbox
from .CompositeWriteBoolean import CompositeWriteBoolean
from .CompositeWriteNumber import CompositeWriteNumber
from .ConstantNumber import ConstantNumber
from .ConstantOn import ConstantOn
from .Delta import Delta
from .Divide import Divide
from .Equal import Equal
from .GreaterThan import GreaterThan
from .JKFlipFlop import JKFlipFlop
from .LessThan import LessThan
from .LuaScript import LuaScript
from .MemoryRegister import MemoryRegister
from .Modulo import Modulo
from .Multiply import Multiply
from .NAND import NAND
from .NOR import NOR
from .NOT import NOT
from .NumberToCompositeBinary import NumberToCompositeBinary
from .NumericalJunction import NumericalJunction
from .NumericalSwitchbox import NumericalSwitchbox
from .OR import OR
from .PIDController import PIDController
from .PIDControllerAdvanced import PIDControllerAdvanced
from .PropertyDropdown import PropertyDropdown
from .PropertyNumber import PropertyNumber
from .PropertySlider import PropertySlider
from .PropertyText import PropertyText
from .PropertyToggle import PropertyToggle
from .Pulse import Pulse
from .PushToToggle import PushToToggle
from .SRLatch import SRLatch
from .Subtract import Subtract
from .Threshold import Threshold
from .TimerRTF import TimerRTF
from .TimerRTO import TimerRTO
from .TimerTOF import TimerTOF
from .TimerTON import TimerTON
from .TooltipBoolean import TooltipBoolean
from .TooltipNumber import TooltipNumber
from .UpDownCounter import UpDownCounter
from .VideoSwitchbox import VideoSwitchbox
from .XOR import XOR



import sw_mc_lib.Components

from .Component import Component
from .Input import Input
from .Microcontroller import Microcontroller
from .Node import Node, NodePosition
from .NumberProperty import NumberProperty
from .Position import Position
from .Types import (
    ComponentType,
    NodeMode,
    PulseMode,
    SignalType,
    TimerUnit,
    TooltipMode,
)
from .XMLFormatter import format
from .XMLParser import parse



import unittest

from sw_mc_lib.NumberProperty import NumberProperty, string_to_sw_float
from sw_mc_lib.XMLParser import XMLParserElement


class TestNumberProperty(unittest.TestCase):
    def test_from_xml(self) -> None:
        elem: XMLParserElement = XMLParserElement("prop")
        expected: NumberProperty = NumberProperty("0", "prop")
        self.assertEqual(NumberProperty.from_xml(elem), expected)
        elem.attributes["value"] = "1234.5"
        self.assertEqual(NumberProperty.from_xml(elem), expected)
        elem.attributes["text"] = "12345.67"
        expected.text = "12345.67"
        self.assertEqual(NumberProperty.from_xml(elem), expected)

    def test_to_xml(self) -> None:
        elem: NumberProperty = NumberProperty("0")
        expected: XMLParserElement = XMLParserElement(
            "temp", {"text": "0", "value": "0.0"}
        )
        self.assertEqual(elem.to_xml(), expected)
        elem.name = "abc"
        expected.tag = "abc"
        self.assertEqual(elem.to_xml(), expected)
        elem.text = "1234567.1234567"
        expected.attributes["text"] = elem.text
        expected.attributes["value"] = "1234567.125"
        self.assertEqual(elem.to_xml(), expected)

    def test_value_property(self) -> None:
        elem: NumberProperty = NumberProperty("0")
        self.assertEqual(elem.value, 0)
        elem.text = "1234"
        self.assertEqual(elem.value, 1234)
        elem.text = "1234567.1234567"
        self.assertEqual(elem.value, 1234567.125)
        elem.value = 1234
        self.assertEqual(elem.text, "1234")
        elem.value = 1234.567
        self.assertEqual(elem.text, "1234.567")

    def test_string_to_sw_float(self) -> None:
        self.assertEqual(string_to_sw_float("0"), 0)
        self.assertEqual(string_to_sw_float("1234"), 1234)
        self.assertEqual(string_to_sw_float("1234567.1234567"), 1234567.125)
        self.assertEqual(string_to_sw_float("0.123456789"), 0.123457)
        self.assertEqual(string_to_sw_float("1.23456789"), 1.234568)



from __future__ import annotations

from typing import Optional

from .Types import NodeMode, SignalType
from .XMLElement import XMLElement
from .XMLParser import XMLParserElement


class NodePosition(XMLElement):
    """
    A Position for a node. This is distinct from :class:`sw_mc_lib.Position.Position` in that it has x, y and z, and
    that they are integers instead of floats. You probably never need to use y.
    """

    def __init__(self, x: int, y: int, z: int):
        self.x: int = x
        self.y: int = y
        self.z: int = z

    @staticmethod
    def from_xml(element: XMLParserElement) -> NodePosition:
        assert element.tag == "position", f"Invalid node position {element}"
        return NodePosition(
            int(element.attributes.get("x", "0")),
            int(element.attributes.get("y", "0")),
            int(element.attributes.get("z", "0")),
        )

    def to_xml(self) -> XMLParserElement:
        attributes: dict[str, str] = {}
        if self.x != 0:
            attributes["x"] = str(self.x)
        if self.y != 0:
            attributes["y"] = str(self.y)
        if self.z != 0:
            attributes["z"] = str(self.z)
        return XMLParserElement("position", attributes)


class Node(XMLElement):
    """
    A Microcontroller node. Interface between Microcontroller and the rest of the Vehicle.
    """

    def __init__(
        self,
        node_id: int,
        component_id: int,
        label: str,
        mode: int,
        type: int,
        description: str,
        position: Optional[NodePosition] = None,
    ):
        self.node_id: int = node_id
        self.component_id: int = component_id
        self.label: str = label
        self.mode: NodeMode = NodeMode(mode)
        self.type: SignalType = SignalType(type)
        self.description: str = description
        self.position: Optional[NodePosition] = position

    @staticmethod
    def from_xml(element: XMLParserElement) -> Node:
        assert element.tag == "n", f"Invalid node {element}"
        node = element.children[0]
        assert node.tag == "node", f"Invalid node {element}"
        position: Optional[NodePosition] = None
        if node.children:
            position = NodePosition.from_xml(node.children[0])
        return Node(
            int(element.attributes.get("id", "1")),
            int(element.attributes.get("component_id", "1")),
            node.attributes.get("label", ""),
            int(node.attributes.get("mode", "0")),
            int(node.attributes.get("type", "0")),
            node.attributes.get("description", ""),
            position,
        )

    def to_xml(self) -> XMLParserElement:
        node_attributes: dict[str, str] = {
            "label": self.label,
            "mode": str(self.mode.value),
            "type": str(self.type.value),
            "description": self.description,
        }
        node_element: XMLParserElement = XMLParserElement("node", node_attributes)
        if self.position:
            node_element.children.append(self.position.to_xml())
        return XMLParserElement(
            "n",
            {"id": str(self.node_id), "component_id": str(self.component_id)},
            [node_element],
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Generator

from .util import generic_str
from .XMLParser import XMLParserElement


class XMLElement(ABC):
    """
    An element that can be deserialized from and serialized to an XML element
    """

    @abstractmethod
    def to_xml(self) -> XMLParserElement:
        """
        Serialize the element to an XML element.

        :return: The XML element
        """

    @staticmethod
    @abstractmethod
    def from_xml(element: XMLParserElement) -> XMLElement:
        """
        Deserialize the element from an XML element.

        :param element: XML element
        :return: Deserialized element
        """

    def __repr__(self) -> str:
        return generic_str(self)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return all(
            callable(getattr(self, i)) or getattr(self, i) == getattr(other, i)
            for i in self.__dir()
        )

    def __dir(self) -> Generator[str, None, None]:
        return (i for i in dir(self) if not i.startswith("__"))



from __future__ import annotations

from .XMLElement import XMLElement
from .XMLParser import XMLParserElement


class Position(XMLElement):
    """
    Position of a Component. Used in 0.25 intervals.
    """

    def __init__(self, x: float = 0.0, y: float = 0.0):
        self.x: float = x
        self.y: float = y

    @staticmethod
    def from_xml(element: XMLParserElement) -> Position:
        assert element.tag == "pos", f"Invalid node position {element}"
        return Position(
            float(element.attributes.get("x", "0")),
            float(element.attributes.get("y", "0")),
        )

    def to_xml(self) -> XMLParserElement:
        attributes: dict[str, str] = {}
        if self.x != 0:
            attributes["x"] = str(self.x)
        if self.y != 0:
            attributes["y"] = str(self.y)
        return XMLParserElement("pos", attributes)



from __future__ import annotations

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class ConstantOn(Component):
    """
    Outputs a constant on signal.
    """

    def __init__(self, component_id: int, position: Position):
        super().__init__(ComponentType.ConstantOn, component_id, position, 0.5)

    @staticmethod
    def from_xml(element: XMLParserElement) -> ConstantOn:
        assert element.tag == "c", f"invalid ConstantOn {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.ConstantOn.value
        ), f"Not an ConstantOn {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, _ = ConstantOn._basic_in_parsing(obj)
        return ConstantOn(component_id, position)

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        return {}, self._pos_in_to_xml()



from typing import Optional

from .XMLParser import XMLParserElement


def _escape_string(to_escape: str) -> str:
    """
    Escapes a string so that it results in valid xml. To be used like following:

    `abc<>"def\n"` -> `'abc&lt;&gt;"def\n"'`

    :param to_escape: The string that needs to be escaped
    :return: Escaped string
    """
    to_escape = to_escape.replace("&", "&amp;")
    to_escape = to_escape.replace("<", "&lt;")
    to_escape = to_escape.replace(">", "&gt;")
    if to_escape.count('"') > to_escape.count("'"):
        to_escape = to_escape.replace("'", "&apos;")
        to_escape = f"'{to_escape}'"
    else:
        to_escape = to_escape.replace('"', "&quot;")
        to_escape = f'"{to_escape}"'
    to_escape = to_escape.replace("\r\n", "\n")
    return to_escape


def _indent(to_indent: str, indentation_character: str, line_breaks: str) -> str:
    """
    Indents the string by one tab and adds a newline at the end. Will only indent lines with `\r\n`

    `<node/>` -> `\t<node/>\n`

    :param to_indent: XML block to indent
    :param indentation_character: The character(s) to use for indentation
    :param line_breaks: Either line breaks or empty string
    :return: Indented text
    """
    return (
        "\r\n".join(
            indentation_character + line for line in to_indent.rstrip().split("\r\n")
        )
        + line_breaks
    )


def _inner_format(element: XMLParserElement, indentation: str, line_breaks: str) -> str:
    xml: str = f"<{element.tag}"
    for name, value in element.attributes.items():
        xml += f" {name}={_escape_string(value)}"
    if element.children:
        xml += f">{line_breaks}"
        for child in element.children:
            xml += _indent(
                _inner_format(child, indentation, line_breaks), indentation, line_breaks
            )
        xml += f"</{element.tag}>{line_breaks}"
    else:
        xml += f"/>{line_breaks}"
    return xml


def format(
    element: XMLParserElement, indentation: Optional[str] = "\t", header: bool = False
) -> str:
    """
    Formats a XMLParserElement into a xml string. If indentation is None, there will be no line breaks or indentation.
    If header is true, it will add a xml declaration at the top of the document.

    :param element: The XML element to format
    :param indentation: The indentation character to use, `None` will not indent at all
    :param header: Whether to include the XML Header
    :return: XML string
    """
    line_breaks: str = "\r\n" if indentation is not None else ""
    indentation_character = indentation or ""
    xml: str = f'<?xml version="1.0" encoding="UTF-8"?>{line_breaks}' if header else ""
    xml += _inner_format(element, indentation_character, line_breaks)
    return xml.replace("\r\n", "\n")



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.Input import Input
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLParser import XMLParserElement


class BooleanFunction8In(Component):
    """
    Evaluates a logical expression with up to 8 input variables and outputs the result.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        function: str,
        x_input: Optional[Input] = None,
        y_input: Optional[Input] = None,
        z_input: Optional[Input] = None,
        w_input: Optional[Input] = None,
        a_input: Optional[Input] = None,
        b_input: Optional[Input] = None,
        c_input: Optional[Input] = None,
        d_input: Optional[Input] = None,
    ):
        super().__init__(ComponentType.BooleanFunction8In, component_id, position, 2.25)
        self.function: str = function
        self.x_input: Optional[Input] = x_input
        self.y_input: Optional[Input] = y_input
        self.z_input: Optional[Input] = z_input
        self.w_input: Optional[Input] = w_input
        self.a_input: Optional[Input] = a_input
        self.b_input: Optional[Input] = b_input
        self.c_input: Optional[Input] = c_input
        self.d_input: Optional[Input] = d_input

    @staticmethod
    def from_xml(element: XMLParserElement) -> BooleanFunction8In:
        assert element.tag == "c", f"invalid BooleanFunction8In {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.BooleanFunction8In.value
        ), f"Not an BooleanFunction8In {element}"
        obj: XMLParserElement = element.children[0]
        (
            component_id,
            position,
            inputs,
            _,
        ) = BooleanFunction8In._basic_in_parsing(obj)
        function: str = obj.attributes.get("e", "")
        return BooleanFunction8In(
            component_id,
            position,
            function,
            inputs.get("1"),
            inputs.get("2"),
            inputs.get("3"),
            inputs.get("4"),
            inputs.get("5"),
            inputs.get("6"),
            inputs.get("7"),
            inputs.get("8"),
        )

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {"e": self.function}
        children: list[XMLParserElement] = self._pos_in_to_xml(
            self.x_input,
            self.y_input,
            self.z_input,
            self.w_input,
            self.a_input,
            self.b_input,
            self.c_input,
            self.d_input,
        )
        return attributes, children



from __future__ import annotations

from typing import Optional

from sw_mc_lib.Component import INNER_TO_XML_RESULT, Component
from sw_mc_lib.NumberProperty import NumberProperty
from sw_mc_lib.Position import Position
from sw_mc_lib.Types import ComponentType
from sw_mc_lib.XMLElement import XMLElement
from sw_mc_lib.XMLParser import XMLParserElement


class DropDownOption(XMLElement):
    """
    A single Option for :class:`sw_mc_lib.Components.PropertyDropDown.PropertyDropDown`
    """

    def __init__(self, label: str, value_property: Optional[NumberProperty]):
        self.label: str = label
        self.value_property: NumberProperty = value_property or NumberProperty("0", "v")

    @staticmethod
    def from_xml(element: XMLParserElement) -> DropDownOption:
        assert element.tag == "i", f"invalid DropDownOption {element}"
        label: str = element.attributes.get("l", "")
        value_property: Optional[NumberProperty] = None
        for child in element.children:
            assert child.tag == "v", f"invalid value field for DropDownOption {element}"
            value_property = NumberProperty.from_xml(child)
        return DropDownOption(label, value_property)

    def to_xml(self) -> XMLParserElement:
        return XMLParserElement("i", {"l": self.label}, [self.value_property.to_xml()])


class PropertyDropdown(Component):
    """
    Adds a custom dropdown list that will be seen on the microcontroller's property panel when placed on a vehicle.
    """

    def __init__(
        self,
        component_id: int,
        position: Position,
        selected_property: int = 0,
        options: Optional[list[DropDownOption]] = None,
    ):
        super().__init__(ComponentType.PropertyDropdown, component_id, position, 0.5)
        self.selected_property: int = selected_property
        self.options: list[DropDownOption] = options or []

    @staticmethod
    def from_xml(element: XMLParserElement) -> PropertyDropdown:
        assert element.tag == "c", f"invalid PropertyDropdown {element}"
        assert element.attributes.get("type", "0") == str(
            ComponentType.PropertyDropdown.value
        ), f"Not an PropertyDropdown {element}"
        obj: XMLParserElement = element.children[0]
        component_id, position, _, _ = PropertyDropdown._basic_in_parsing(obj)
        selected_property: int = int(obj.attributes.get("i", "0"))
        options: list[DropDownOption] = []
        for child in obj.children:
            if child.tag == "items":
                for entry in child.children:
                    options.append(DropDownOption.from_xml(entry))
        return PropertyDropdown(component_id, position, selected_property, options)

    def _inner_to_xml(self) -> INNER_TO_XML_RESULT:
        attributes: dict[str, str] = {"i": str(self.selected_property)}
        children: list[XMLParserElement] = self._pos_in_to_xml()
        items: XMLParserElement = XMLParserElement("items")
        for option in self.options:
            items.children.append(option.to_xml())
        children.append(items)
        return attributes, children
